/**
 * @(#)sidt2.cpp 2002/02/25
 *
 * @author Ice
 * @version 1.00
 *
 * ICE Team free software group
 *
 * This file is part of TSID2.
 * See README for copyright notice.
 *
 * note: some comments are in the style of Javadoc of Java language!
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 */

#ifndef _SIDT2
#define _SIDT2

#include "sidt2.h"

#include <string.h>
#include <stdio.h>
#include <iostream.h>
#include <fstream.h>

#ifdef HAVE_MSWINDOWS
# define snprintf _snprintf
#endif

/**
 * Construnct SIDT2 using the passed information
 * All the passed path separator must be as ANSI
 * The special absSidFileName #/# is to be used when you did want
 * to save the old file name
 *
 * @param MD5 the MD5 fingerprinter of this file
 * @param absSidFileName the absolute sid file path and name
 * @param absBasePath the base path where there are the sidt2 files
 */
SIDT2::SIDT2(const char *MD5, const char *absSidFileName, const char *absBasePath) {
  PDEBUG2("Construct the SIDT2 with "<<MD5<<" " << absSidFileName << " " << absBasePath << "\n");

  versionString="sidt2 1.00 by Ice Team Free Software Group (C) 2000-2002";

  lastError=NO_SIDT2_ERROR;      // no error
  sidtIndex=-1;                  // engine not initialized
  sidtStatus=true;
  SIDT2::MD5=NULL;
  SIDT2::absSidFileName=NULL;
  SIDT2::absBasePath=NULL;
  absFileName=NULL;
  oldAbsSidPath=NULL;
  oldSidName=NULL;
  database=NULL;

  // test for MD5 present
  if (MD5 == NULL) {
    PDEBUG2("MD5 not present\n");
    lastError=NO_MD5_ERROR;      // no MD5 for this song
    sidtStatus=false;            // tsid2 cannot continue
    return;
  }

  // test for path present
  if (absSidFileName == NULL) {
    PDEBUG2("absSidFileName not present\n");
    lastError=NO_FILE_ERROR;     // no file name for this song
    sidtStatus=false;            // tsid2 cannot continue
    return;
  }

  // test for base present
  if (absBasePath == NULL) {
    PDEBUG2("absBasePath not present\n");
    lastError=NO_BASE_ERROR;     // no base path for the stored information
    sidtStatus=false;            // tsid2 cannot continue
    return;
  }

  // make a phisical copy of the passed argument
  SIDT2::MD5=new char[strlen(MD5)+1];
  SIDT2::absSidFileName=new char[strlen(absSidFileName)+1];
  SIDT2::absBasePath=new char[strlen(absBasePath)+1];

  if ((SIDT2::absSidFileName == NULL) || (SIDT2::MD5 == NULL) || (SIDT2::absBasePath == NULL)) {
    PDEBUG2("Error: no memory\n");
    lastError=NO_MEMORY;            // no more memory
    sidtStatus=false;               // sidt cannot go away
    return;
  }
  strcpy(SIDT2::MD5, MD5);
  strcpy(SIDT2::absSidFileName, absSidFileName);
  strcpy(SIDT2::absBasePath, absBasePath);

  // create the complete name of the MD5 files
  absFileName=new char[strlen(absBasePath)+strlen(MD5)+MD5_EXT_LEN+2];
  if (absFileName == NULL) {
    PDEBUG2("Error: no memory\n");
    lastError=NO_MEMORY;            // no more memory
    sidtStatus=false;               // sidt cannot go away
    return;
  }
  sprintf(absFileName, "%s/%s%s", absBasePath, MD5, MD5_EXT);
  // erase memory locations
  memset(&sidt, 0, sizeof(sidt));

  PDEBUG2("absFileName=" << absFileName << endl);
}

/**
 * Destruct the SIDT2
 */
SIDT2::~SIDT2() {
  PDEBUG2("Destruct SIDT2\n");
  if (database != NULL) ini_close(database);

  if (MD5 != NULL)            delete[] MD5;
  if (absSidFileName != NULL) delete[] absSidFileName;
  if (absBasePath != NULL)    delete[] absBasePath;
  if (absFileName != NULL)    delete[] absFileName;
  if (oldAbsSidPath != NULL)  delete[] oldAbsSidPath;
  if (oldSidName != NULL)     delete[] oldSidName;
}

/**
 * Add the passed information to the vector. The vector will be ordinate.
 *
 * @param tune the number of sub-tune to update or add
 * @param second the number of second to add
 * @param number the number of time to add (default=1)
 * @param length the length of the tune (default=0)
 */
void SIDT2::add(unsigned int tune, unsigned long second, unsigned long number, unsigned long length) {
  PDEBUG2("add #"<< tune << " seconds=" << second << " number=" << number << " length=" << length << endl);
  // scan all the vector
  for (int i=0; i<MAX_SONGS; i++) {
    if ((sidt[i].tune==tune) || (sidt[i].tune==0)) {
      PDEBUG2("Update field for tune #" << tune << endl);
      sidt[i].tune=tune;
      sidt[i].second+=second;
      sidt[i].number+=number;
      if (length!=0) {           // assume this length the right one
        sidt[i].length=length;
      }
      break;
    }
    if (tune<sidt[i].tune) {

      PDEBUG2("Creating a free space in record\n");

      // create a space for the new song
      for (int j=MAX_SONGS-1; j>i; j--) {
        sidt[j]=sidt[j-1];
      }
      PDEBUG2("Add field for tune #" << tune << endl);
      sidt[i].tune=tune;
      sidt[i].second=second;
      sidt[i].number=number;
      sidt[i].length=length;
      break;
    }
  }
}

/**
 * Add the passed struct of information to the vector.
 * The vector will be ordinate.
 *
 * @param rec the record of information
 */
void SIDT2::add(struct sidtRecord rec) {
  add(rec.tune, rec.second, rec.number, rec.length);
}

/**
 * Read the record from a file if it's present. The readed record id add to
 * the actual record.
 *
 * @return true if operation is ok
 */
bool SIDT2::read() {
  PDEBUG2("Read the SIDT2 file\n");
  // test if the files exist

  { // open the file
    ifstream file(absFileName, ios::in);

    if (!file) {
      // nothing to do
      PDEBUG2("Nothing to do: file not opened\n");
      return true;
    }
  }

  // test if we need to open the database
  if (database == NULL) {
    database=ini_open(absFileName,"w", NULL);
    if (!database) {
      PDEBUG2("Error opening the INI database for the " << absFileName);
      lastError=INI_OPEN_ERROR;
      return false;
    }
    if (ini_listDelims(database, " ")!=0) {
      PDEBUG2("Error setting delimiters in the INI database");
      lastError=INI_DELIMS_ERROR;
      return false;
    }
  }

  ini_locateHeading(database, "STATS");

  char key[5]="";
  int time;
  int listen;
  int length;

  //now scan for all the possible tunes key
  for (int i=0; i<MAX_SONGS; i++) {
    sprintf(key, "%d", i);
    ini_locateKey(database, key);
    if (ini_dataLength(database)!=-1) {
      // this key is present: add to the vector
      ini_readInt(database, &time);
      ini_readInt(database, &listen);
      ini_readInt(database, &length);
      add((unsigned int) i, (unsigned int) time, (unsigned int) listen, (unsigned int) length);
    }
  }

  // now read the sid path/name
  int len;

  ini_locateHeading(database, "TUNE");

  ini_locateKey(database, "path");
  if ((len=ini_dataLength(database))!=-1) {
    // key present
    oldAbsSidPath=new char[len+1];
    if (oldAbsSidPath==NULL) {
      lastError=NO_MEMORY;            // no more memory
      sidtStatus=false;               // sidt cannot go away
      return false;
    }

    ini_readString(database, oldAbsSidPath, len+1);
  }

  ini_locateKey(database, "name");
  if ((len=ini_dataLength(database))!=-1) {
    // key present
    oldSidName=new char[len+1];
    if (oldSidName==NULL) {
      lastError=NO_MEMORY;            // no more memory
      sidtStatus=false;               // sidt cannot go away
      return false;
    }

    ini_readString(database, oldSidName, len+1);
  }

  return true;
}

/**
 * Write the sidt records in the file
 *
 * @return true if operation ok
 */
bool SIDT2::write() {
#define DIM 80
  PDEBUG2("Write the SIDT2 file\n");
  char str[DIM];

  // test if we need to open the database
  if (database == NULL) {
    database=ini_open(absFileName,"w", NULL);
    if (!database) {
      PDEBUG2("Error opening the INI database for the " << absFileName);
      lastError=INI_OPEN_ERROR;
      return false;
    }
    if (ini_listDelims(database, " ")!=0) {
      PDEBUG2("Error setting delimiters in the INI database");
      lastError=INI_DELIMS_ERROR;
      return false;
    }
  }

  // delete old value
  ini_locateHeading(database, "STATS");
  ini_deleteHeading(database);

  // insert new one
  ini_locateHeading(database, "STATS");

  // insert all the subtunes
  for (int i=0; i<MAX_SONGS; i++) {
    if (sidt[i].tune==0) break;
    snprintf(str, (size_t)DIM, "%d", sidt[i].tune);
    ini_locateKey(database, str);
    snprintf(str, (size_t)DIM, "%lu %lu %lu\n", sidt[i].second, sidt[i].number, sidt[i].length);
    ini_writeString(database, str);
  }

  // separate name from path
  int length;                                        // length of result string
  char *pnt=&absSidFileName[strlen(absSidFileName)]; // pointer to the end of string

  while (*pnt!='/') {                                // look for slash
    pnt--;                                           // decrement pointer
    if (pnt<=absSidFileName) break;                  // avoid endless loop
  }

  char *result=new char[(length=(pnt-absSidFileName))+1];
  if (result==NULL) {
      lastError=NO_MEMORY;                           // no memory available
      sidtStatus=false;
      return false;
  }
  strncpy(result, absSidFileName, length);

  result[length]='\0';                               // terminate the string

  PDEBUG2("BeforeSlash: \""<< result << "\"\n");

  // delete old TUNE info
  ini_locateHeading(database, "TUNE");
  ini_deleteHeading(database);

  // insert new one
  ini_locateHeading(database, "TUNE");

  if (strcmp(absSidFileName, "#/#")==0) {
    ini_locateKey(database, "path");
    if (strlen(oldAbsSidPath)==0) ini_writeString(database, "./");
    else ini_writeString(database, oldAbsSidPath);

    ini_locateKey(database, "name");
    ini_writeString(database, oldSidName);
  } else {

      ini_locateKey(database, "path");
      if (strlen(result)==0) ini_writeString(database, "./");
      else ini_writeString(database, result);

      ini_locateKey(database, "name");
      ini_writeString(database, &absSidFileName[length+1]);
    }

  ini_flush(database);
  return true;
}

/**
 * Shift the sub songs from one by some step
 * This is used when a new sub song is added in a sid file
 *
 * @param step the shift step
 * @param from the number of subtunes from being to shift
 */
void SIDT2::shift(unsigned int step, unsigned int from) {
  PDEBUG2("Shift from song "<< from << " of " << step << endl);

   // scan all the vector
  for (int i=0; i<MAX_SONGS; i++) {
    if (sidt[i].tune==0) {                         // songs finished?
      break;                                       // so, exit
    }
    if (sidt[i].tune>=from) {                      // is a element to shift?
      PDEBUG2("Shift song "<< sidt[i].tune << endl);
      sidt[i].tune+=step;                          // update song number
    }
  }
}

/**
 * Swap the position of two tunes
 *
 * @param tune1 the first tune
 * @param tune2 the second tune
 */
void SIDT2::swap(unsigned int tune1, unsigned int tune2) {
  unsigned int tmp;
  unsigned long tmp2;
  sidtRecord tmp3;

  int i;

  int pos1=-1;
  int pos2=-1;

  PDEBUG2("Change position of song "<< tune1 << " and "<< tune2 << endl);

  // if equal we don't change anything
  if (tune2==tune1) return;

  // song1 must be before song2
  if (tune2<tune1) {
    tmp=tune2;
    tune2=tune1;
    tune1=tmp;
  }

  // scan all the vector
  for (i=0; i<MAX_SONGS; i++) {
    if (sidt[i].tune==0) break;
    if (sidt[i].tune==tune1) pos1=i;
    if (sidt[i].tune==tune2) pos2=i;
  }

  // if the elements are not present we exit
  if ((pos1==-1) && (pos2==-1)) return;

  // best case: we change the elements positions
  if ((pos1!=-1) && (pos2!=-1)) {
    tmp2=sidt[pos2].second;
    sidt[pos2].second=sidt[pos1].second;
    sidt[pos1].second=tmp2;

    tmp2=sidt[pos2].number;
    sidt[pos2].number=sidt[pos1].number;
    sidt[pos1].number=tmp2;

    tmp2=sidt[pos2].length;
    sidt[pos2].length=sidt[pos1].length;
    sidt[pos1].length=tmp2;
    return;
  }

  // pos2 is present, but not pos1
  if (pos2!=-1) {
    tmp3=sidt[pos2];
    for (i=pos2-1; i>=0; i--) {
      // have we reach a position below song1?
      if (sidt[i].tune<tune1) {
        sidt[i+1]=tmp3;
        sidt[i+1].tune=tune1;
        return;
      }
      sidt[i+1]=sidt[i];
    }
    sidt[1]=sidt[0];
    sidt[0]=tmp3;
    sidt[0].tune=tune1;
    return;
  }

  // pos 1 is present, but not pos2
  tmp3=sidt[pos1];
  for (i=pos1+1; i<MAX_SONGS; i++) {
    if ((sidt[i].tune>tune2) || (sidt[i].tune==0) ) {
      sidt[i-1]=tmp3;
      sidt[i-1].tune=tune2;
      return;
    }
    sidt[i-1]=sidt[i];
  }
}

/**
 * Remove the field for the passed tune from memory
 *
 * @param song the number of tune to erase
 */
void SIDT2::remove(unsigned int tune) {
  PDEBUG2("Remove song " << tune << " from the memory\n");

  // scan all the vector
  for (int i=0; i<MAX_SONGS; i++) {
    if (sidt[i].tune==0) break;

    if ((sidt[i].tune==tune)) {
      // shift all the other songs for eliminating this
      for (int j=i; j<MAX_SONGS-1; j++) {
        sidt[j]=sidt[j+1];
      }
      sidt[MAX_SONGS-1].tune=0;
      sidt[MAX_SONGS-1].number=0;
      sidt[MAX_SONGS-1].second=0;
      sidt[MAX_SONGS-1].length=0;
      break;
    }
  }
}

/**
 * Init the engin for getting the next tune information
 */
void SIDT2::initForGet() {
  sidtIndex=0;   // go to the first element
}

/**
 * Get the next available tune information from the sidt2 record
 *
 * @param tune the returned tune number
 * @param second the returned number of seconds
 * @param number the returned number of time
 * @param length the returned tune length (0=unknown)
 * @return true if the record is present
 */
bool SIDT2::getNext(unsigned int *tune, unsigned long *second,
                   unsigned long *number, unsigned long *length) {
  if (sidtIndex==-1) {
    return false;            // the engine is not initialized
  }

  if (sidt[sidtIndex].tune==0) {
    return false;            // no other songs present
  }

  *tune=sidt[sidtIndex].tune;
  *second=sidt[sidtIndex].second;
  *number=sidt[sidtIndex].number;
  *length=sidt[sidtIndex].length;

  sidtIndex++;
  return true;
}

/**
 * Get the next available information as a record
 *
 * @param rec the returned record of information
 * @return true if the record is present
 */
bool SIDT2::getNext(struct sidtRecord *rec) {
  return getNext(&rec->tune, &rec->second, &rec->number, &rec->length);
}

/**
 * Get the information for the passed tune if present
 *
 * @param tune the tune to look for
 * @param rec the returned record of information
 * @return true if the tune is present
 */
bool SIDT2::getTune(unsigned int tune, struct sidtRecord *rec) {
  PDEBUG2("Get information of tune #" << tune << endl);

  // scan all the vector
  for (int i=0; i<MAX_SONGS; i++) {
    if (sidt[i].tune==0) return false;

    if ((sidt[i].tune==tune)) {
      // read information
      rec->tune=sidt[i].tune;
      rec->number=sidt[i].number;
      rec->second=sidt[i].second;
      rec->length=sidt[i].length;
      return true;
    }
  }
  return false;
}

/**
 * Get a copy of the readed sid file path
 *
 * @return a copy of the sid file path
 */
char *SIDT2::getReadSidPath() {
  PDEBUG2("Get the read sid path: ");
  char *tmp;

  if (oldAbsSidPath==NULL) return NULL;

  tmp=new char[strlen(oldAbsSidPath)+1];
  if (tmp == NULL) {
    PDEBUG2("No memory available\n");
    lastError=NO_MEMORY;            // no more memory
    sidtStatus=false;               // sidt cannot go away
    return NULL;
  }
  strcpy(tmp, oldAbsSidPath);
  PDEBUG2(tmp << endl);
  return tmp;
}

/**
 * Get a copy of the readed sid file name
 *
 * @return a copy of the sid file name
 */
char *SIDT2::getReadSidName() {
  PDEBUG2("Get the read sid name: ");
  char *tmp;

  if (oldSidName==NULL) return NULL;

  tmp=new char[strlen(oldSidName)+1];
  if (tmp == NULL) {
    lastError=NO_MEMORY;            // no more memory
    sidtStatus=false;               // sidt cannot go away
    return NULL;
  }
  strcpy(tmp, oldSidName);
  PDEBUG2(tmp << endl);
  return tmp;
}
    
/**
 * Get a copy of the MD5 fingerprinter of this file
 *
 * @return a copy of the MD5 of this file
 */
char *SIDT2::getMD5() {
  PDEBUG2("Get MD5: " << MD5 << endl);
  
  if (MD5==NULL) return NULL;

  char *tmp=new char[strlen(MD5)+1];
  if (tmp == NULL) {
    lastError=NO_MEMORY;            // no more memory
    sidtStatus=false;               // sidt cannot go away
    return NULL;
  }
  strcpy(tmp, MD5);
  return tmp;
}

#endif // _SIDT2.CPP


